import cv2
import numpy as np

class Recorder:
    def __init__(self, output_path: str, output_dims: tuple, fps: int):
        fourcc = cv2.VideoWriter_fourcc('M','J','P','G')
        self.output_dims = output_dims
        self.fps = fps
        self.record_out = cv2.VideoWriter(output_path, fourcc, fps, self.output_dims)

    def write(self, frame: np.ndarray):
        self.record_out.write(frame)

    def close(self):
        self.record_out.release()


import cv2
import numpy as np

def get_scaled_dims(width: int, height: int, scale_factor) -> (int, int):
    return int(width * scale_factor), int(height * scale_factor)

def scale_img(frame: np.ndarray, scale_factor) -> np.ndarray:
    height, width = frame.shape[:2]
    downsized_width, downsized_height = get_scaled_dims(width, height, scale_factor)

    if scale_factor == 1.0:    
        return frame
    elif scale_factor > 0.0 and scale_factor < 1.0:
        return cv2.resize(
            frame,
            (downsized_width, downsized_height),
            interpolation=cv2.INTER_AREA)
    elif scale_factor > 1.0:
        return cv2.resize(
            frame,
            (downsized_width, downsized_height),
            interpolation=cv2.INTER_LINEAR)
    else:
        raise Exception(f"Invalid scale_factor={scale_factor}. Expected scale_factor > 0.")

def get_grayscale(frame: np.ndarray) -> np.ndarray:
    return cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)


from typing import List
import cv2
import numpy as np
from logger import logger
from common_utils.check_utils import check_type, check_file_exists, check_value
from common_utils.image_utils import concat_images

from ...streamer import Streamer, DualStreamer, StreamerObject
from ...recorder import Recorder

class VideoConcatenator:
    def __init__(
        self, src0: str, src1: str, output_path: str,
        stream_mode: str='mono', default_fps: int=25,
        src0_scale_factor: float=1.0, src1_scale_factor: float=1.0, output_scale_factor: float=1.0,
        orientation: int=1
    ):
        """
        orientation
        0: horizontal
        1: vertical
        """
        
        # Create Streamers
        self.streamer0 = self.create_streamer(src=src0, mode=stream_mode, scale_factor=src0_scale_factor, verbose=True)
        self.streamer1 = self.create_streamer(src=src1, mode=stream_mode, scale_factor=src1_scale_factor, verbose=True)

        fps0, fps1 = self.streamer0.get_fps(), self.streamer1.get_fps()
        self.fps = fps0 if fps0 == fps1 else default_fps

        # Recorder
        self.recorder = None

        # Other Parameters
        self.output_path = output_path
        self.navigation_step = 10
        self.src0_scale_factor, self.src1_scale_factor = src0_scale_factor, src1_scale_factor
        self.output_scale_factor = output_scale_factor
        self.orientation = orientation

    def create_streamer(self, src: str, mode: str='mono', scale_factor: float=1.0, verbose: bool=False) -> StreamerObject:
        if verbose: logger.info(f"Creating Streamer for src={src}")
        check_value(item=mode, valid_value_list=['mono', 'dual'])
        check_file_exists(src)
        if mode == 'mono':
            streamer = Streamer(src=src, scale_factor=scale_factor)
        elif mode == 'dual':
            streamer = DualStreamer(src=src, scale_factor=scale_factor, direction=0)
        else:
            raise Exception
        return streamer

    def write_and_display_result(self, result: np.ndarray) -> bool:
        break_flag = False
        result_h, result_w = result.shape[:2]

        if self.recorder is None:
            self.recorder = Recorder(output_path=self.output_path, output_dims=(result_w, result_h), fps=self.fps)
        
        self.recorder.write(result)
        window_h, window_w = int(result_h * self.output_scale_factor), int(result_w * self.output_scale_factor)
        window_name = 'test'
        cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)
        cv2.resizeWindow(window_name, (window_w, window_h))
        cv2.imshow(window_name, result)
        key = cv2.waitKey(1) & 0xFF
        if key == ord('q'):
            break_flag = True
        elif key == ord('a'):
            self.streamer0.rewind(self.navigation_step)
            self.streamer1.rewind(self.navigation_step)
        elif key == ord('d'):
            self.streamer0.fastforward(self.navigation_step)
            self.streamer1.fastforward(self.navigation_step)
        elif key == ord('w'):
            self.navigation_step = self.navigation_step * 10 if self.navigation_step < 100 else 100
        elif key == ord('s'):
            self.navigation_step = self.navigation_step / 10 if self.navigation_step > 1 else 100

        return break_flag

    def run(self):
        # Main Loop
        logger.info("Main Loop Start")
        while self.streamer0.is_playing() and self.streamer1.is_playing():
            logger.info(f"Frame {self.streamer0.get_progress_ratio_string()} | {self.streamer1.get_progress_ratio_string()}")
            if not self.streamer0.is_open() or not self.streamer1.is_open():
                logger.warning(f"Streamer is not open. Terminating...")
                break
            frame0, frame1 = self.streamer0.get_frame(), self.streamer1.get_frame()
            if frame0 is None or frame1 is None:
                logger.warning(f"Frame is None. Terminating...")
                break

            if self.orientation == 0 and frame0.shape[0] != frame1.shape[0]:
                logger.error(f"frame0.shape[0] == {frame0.shape[0]} != {frame1.shape[0]} == frame1.shape[0]")
                raise Exception
            elif self.orientation == 1 and frame0.shape[1] != frame1.shape[1]:
                logger.error(f"frame0.shape[1] == {frame0.shape[1]} != {frame1.shape[1]} == frame1.shape[1]")
                raise Exception

            result = concat_images(imga=frame0, imgb=frame1, orientation=self.orientation)

            break_flag = self.write_and_display_result(result=result)
            if break_flag:
                logger.warning(f"Break flag triggered. Terminating...")
                break

        self.recorder.close()
        self.streamer0.close()
        self.streamer1.close()


from setuptools import setup, find_packages
import streamer

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pyclay-streamer',
    version=streamer.__version__,
    description='OpenCV Video Streaming and Recording API',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cm107/streamer",
    author='Clayton Mork',
    author_email='mork.clayton3@gmail.com',
    license='MIT License',
    packages=find_packages(
        where='.',
        include=['streamer*']
    ),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'opencv-python>=4.1.1.26',
        'numpy>=1.17.2',
        'pylint>=2.4.2',
        'pyclay-common_utils>=0.1',
        'pyclay-logger==0.2'
    ],
    python_requires='>=3.6'
)



import cv2
import numpy as np
from abc import ABCMeta, abstractmethod
from ..util import scale_img
from logger import logger
from common_utils.check_utils import check_value
from common_utils.utils import get_class_string

class StreamerObject(metaclass=ABCMeta):
    def __init__(self, src, scale_factor: float=1.0):
        self.src = src
        self.worker = cv2.VideoCapture(src)
        self.full_width = None
        self.full_height = None
        self.width = None
        self.height = None
        self.scale_factor = scale_factor
        self.downsized_width = None
        self.downsized_height = None
    
    @abstractmethod
    def init_dims(self):
        pass

    @abstractmethod
    def get_frames(self):
        pass

    @abstractmethod
    def get_frame(self) -> np.ndarray:
        pass

    @abstractmethod
    def _update_frame_buff(self, frame: np.ndarray) -> np.ndarray:
        pass

    def get_frame_count(self) -> int:
        if type(self.src) is str:
            return int(self.worker.get(7))
        else:
            return -1

    def is_open(self) -> bool:
        return self.worker.isOpened()

    def get_num_frames_read(self) -> int:
        if type(self.src) is str:
            return int(self.worker.get(1))
        else:
            return -1

    def is_playing(self) -> bool:
        if type(self.src) is str:
            return self.get_num_frames_read() < self.get_frame_count()
        else:
            return self.is_open()

    def get_progress_ratio_string(self) -> str:
        return f"{self.get_num_frames_read()}/{self.get_frame_count()}"

    def get_fps(self) -> int:
        return int(self.worker.get(5))

    def goto_frame(self, target_frame_num: int) -> bool:
        if type(self.src) is str:
            success = self.worker.set(1, target_frame_num)
            return success
        else:
            return False

    def is_mono(self) -> bool:
        return isinstance(self, Streamer)

    def is_dual(self) -> bool:
        return isinstance(self, DualStreamer)

    def rewind(self, number_of_frames: int):
        total_frames = self.get_frame_count()
        current_frame_num = self.get_num_frames_read()
        target_frame = current_frame_num - number_of_frames
        target_frame = 0 if target_frame < 0 else target_frame
        success = self.goto_frame(target_frame)
        if not success:
            logger.warning(f'Failed to rewind to {target_frame}/{total_frames}')

    def fastforward(self, number_of_frames: int):
        total_frames = self.get_frame_count()
        current_frame_num = self.get_num_frames_read()
        target_frame = current_frame_num + number_of_frames
        target_frame = total_frames if target_frame > total_frames else target_frame
        success = self.goto_frame(target_frame)
        if not success:
            logger.warning(f'Failed to fastforward to {target_frame}/{total_frames}')

    def sample_frame_shape(self) -> list:
        if type(self.src) is str:
            frame = self.get_frame()
            self.rewind(1)
            return frame.shape
        elif type(self.src) is int:
            frame = self.get_frame()
            return frame.shape
        else:
            raise Exception

    def assert_open(self):
        if not self.is_open():
            logger.error(
                f"{get_class_string(self)} is not open.\n" + \
                "Please check your video source.\n" + \
                f"Attempted to use {self.src}")
            raise Exception

    def close(self):
        self.worker.release()
        cv2.destroyAllWindows()

class Streamer(StreamerObject):
    def __init__(self, src, scale_factor: float=1.0):
        super().__init__(src, scale_factor)
        self.assert_open()
        self.init_dims()
        self.current_frame = None

    def init_dims(self):
        self.full_width = int(self.worker.get(3))
        self.full_height = int(self.worker.get(4))
        self.width = self.full_width
        self.height = self.full_height
        self.downsized_width = int(self.width * self.scale_factor)
        self.downsized_height = int(self.height * self.scale_factor)

    def get_frames(self) -> np.ndarray:
        rec, frame = self.worker.read()
        if rec:
            self._update_frame_buff(frame)
            return scale_img(frame, self.scale_factor)
        else:
            return None

    def get_frame(self) -> np.ndarray:
        return self.get_frames()

    def _update_frame_buff(self, frame: np.ndarray):
        self.current_frame = frame if frame is not None else self.current_frame

class DualStreamer(StreamerObject):
    def __init__(self, src, scale_factor: float=1.0, direction: int=0):
        """
        direction=0 : left-right direction
        direction=1 : top-bottom direction
        """

        super().__init__(src, scale_factor)
        self.assert_open()
        check_value(item=direction, valid_value_list=[0, 1])
        self.direction = direction
        self.init_dims()
        self.current_left_frame = None
        self.current_right_frame = None

    def init_dims(self):
        self.full_width = int(self.worker.get(3))
        self.full_height = int(self.worker.get(4))
        if self.direction == 0: # left-right direction
            self.width = int(self.full_width / 2)
            self.height = self.full_height
        elif self.direction == 1: # top-down direction
            self.width = self.full_width
            self.height = int(self.full_height / 2)
        else:
            logger.error(f"Invalid direction: {self.direction}.")
            logger.error("Expecting direction in [0, 1]")
            raise Exception
        self.downsized_width = int(self.width * self.scale_factor)
        self.downsized_height = int(self.height * self.scale_factor)

    def get_frames(self) -> (np.ndarray, np.ndarray):
        rec, frame = self.worker.read()
        if rec:
            if self.direction == 0:
                left_frame = frame[:,:self.width,:]
                right_frame = frame[:,self.width:,:]
            elif self.direction == 1:
                left_frame = frame[:self.height,:,:]
                right_frame = frame[self.height:,:,:]
            self._update_frame_buff(left_frame, right_frame)
            return scale_img(left_frame, self.scale_factor), scale_img(right_frame, self.scale_factor)
        else:
            return None, None

    def get_frame(self, side: str=None) -> np.ndarray:
        frame0, frame1 = self.get_frames()
        if self.direction == 0: # left-right direction
            use_side = None
            if side is not None:
                check_value(item=side, valid_value_list=['left', 'right'])
                use_side = side
            else:
                use_side = 'left'
            if use_side == 'left':
                return frame0
            elif use_side == 'right':
                return frame1
            else:
                raise Exception
        elif self.direction == 1: # top-down direction
            use_side = None
            if side is not None:
                check_value(item=side, valid_value_list=['top', 'down'])
                use_side = side
            else:
                use_side = 'top'
            if use_side == 'top':
                return frame0
            elif use_side == 'right':
                return frame1
            else:
                raise Exception
        else:
            raise Exception

    def _update_frame_buff(self, left_frame: np.ndarray, right_frame: np.ndarray):
        self.current_left_frame = left_frame if left_frame is not None else self.current_left_frame
        self.current_right_frame = right_frame if right_frame is not None else self.current_right_frame
